package prog8.codegen.cpu6502

import com.github.michaelbull.result.fold
import prog8.ast.IFunctionCall
import prog8.ast.IStatementContainer
import prog8.ast.Program
import prog8.ast.antlr.escape
import prog8.ast.base.*
import prog8.ast.expressions.*
import prog8.ast.statements.*
import prog8.ast.toHex
import prog8.codegen.cpu6502.assignment.AsmAssignTarget
import prog8.codegen.cpu6502.assignment.TargetStorageKind
import prog8.compilerinterface.*
import java.time.LocalDate
import java.time.LocalDateTime
import kotlin.math.absoluteValue


internal class ProgramGen(
    val program: Program,
    val variables: IVariablesAndConsts,
    val options: CompilationOptions,
    val errors: IErrorReporter,
    private val functioncallAsmGen: FunctionCallAsmGen,
    private val asmgen: AsmGen
) {
    private val compTarget = options.compTarget
    private val removals = mutableListOf<Pair<Statement, IStatementContainer>>()
    private val allocation = VariableAllocation(variables, errors)
    private val callGraph = CallGraph(program)
    private val blockVariableInitializers = program.allBlocks.associateWith { it.statements.filterIsInstance<Assignment>() }

    internal fun generate() {
        // variables.dump(program.memsizer)        // TODO

        val allInitializers = blockVariableInitializers.asSequence().flatMap { it.value }
        require(allInitializers.all { it.origin==AssignmentOrigin.VARINIT }) {"all block-level assignments must be a variable initializer"}

        header()
        val allBlocks = program.allBlocks
        if(allBlocks.first().name != "main")
            throw AssemblyError("first block should be 'main'")

        allocation.allocateAllZeropageVariables(options, callGraph)
        if(errors.noErrors())  {
            program.allBlocks.forEach { block2asm(it) }

            for(removal in removals.toList()) {
                removal.second.remove(removal.first)
                removals.remove(removal)
            }

            memorySlabs()
            footer()
        }
    }

    private fun header() {
        val ourName = this.javaClass.name
        val cpu = when(compTarget.machine.cpu) {
            CpuType.CPU6502 -> "6502"
            CpuType.CPU65c02 -> "w65c02"
            else -> "unsupported"
        }

        asmgen.out("; $cpu assembly code for '${program.name}'")
        asmgen.out("; generated by $ourName on ${LocalDateTime.now().withNano(0)}")
        asmgen.out("; assembler syntax is for the 64tasm cross-assembler")
        asmgen.out("; output options: output=${options.output} launcher=${options.launcher} zp=${options.zeropage}")
        asmgen.out("\n.cpu  '$cpu'\n.enc  'none'\n")

        program.actualLoadAddress = program.definedLoadAddress
        if (program.actualLoadAddress == 0u)   // fix load address
            program.actualLoadAddress = if (options.launcher == LauncherType.BASIC)
                compTarget.machine.BASIC_LOAD_ADDRESS else compTarget.machine.RAW_LOAD_ADDRESS

        // the global prog8 variables needed
        val zp = compTarget.machine.zeropage
        asmgen.out("P8ZP_SCRATCH_B1 = ${zp.SCRATCH_B1}")
        asmgen.out("P8ZP_SCRATCH_REG = ${zp.SCRATCH_REG}")
        asmgen.out("P8ZP_SCRATCH_W1 = ${zp.SCRATCH_W1}    ; word")
        asmgen.out("P8ZP_SCRATCH_W2 = ${zp.SCRATCH_W2}    ; word")
        asmgen.out("P8ESTACK_LO = ${compTarget.machine.ESTACK_LO.toHex()}")
        asmgen.out("P8ESTACK_HI = ${compTarget.machine.ESTACK_HI.toHex()}")

        when {
            options.launcher == LauncherType.BASIC -> {
                if (program.actualLoadAddress != options.compTarget.machine.BASIC_LOAD_ADDRESS)
                    throw AssemblyError("BASIC output must have correct load address")
                asmgen.out("; ---- basic program with sys call ----")
                asmgen.out("* = ${program.actualLoadAddress.toHex()}")
                val year = LocalDate.now().year
                asmgen.out("  .word  (+), $year")
                asmgen.out("  .null  $9e, format(' %d ', prog8_entrypoint), $3a, $8f, ' prog8'")
                asmgen.out("+\t.word  0")
                asmgen.out("prog8_entrypoint\t; assembly code starts here\n")
                if(!options.noSysInit)
                    asmgen.out("  jsr  ${compTarget.name}.init_system")
                asmgen.out("  jsr  ${compTarget.name}.init_system_phase2")
            }
            options.output == OutputType.PRG -> {
                asmgen.out("; ---- program without basic sys call ----")
                asmgen.out("* = ${program.actualLoadAddress.toHex()}\n")
                if(!options.noSysInit)
                    asmgen.out("  jsr  ${compTarget.name}.init_system")
                asmgen.out("  jsr  ${compTarget.name}.init_system_phase2")
            }
            options.output == OutputType.RAW -> {
                asmgen.out("; ---- raw assembler program ----")
                asmgen.out("* = ${program.actualLoadAddress.toHex()}\n")
            }
        }

        if(options.zeropage !in arrayOf(ZeropageType.BASICSAFE, ZeropageType.DONTUSE)) {
            asmgen.out("""
                ; zeropage is clobbered so we need to reset the machine at exit
                lda  #>sys.reset_system
                pha
                lda  #<sys.reset_system
                pha""")
        }

        // make sure that on the cx16 and c64, basic rom is banked in again when we exit the program
        when(compTarget.name) {
            "cx16" -> {
                if(options.floats)
                    asmgen.out("  lda  #4 |  sta  $01")    // to use floats, make sure Basic rom is banked in
                asmgen.out("  jsr  main.start |  lda  #4 |  sta  $01 |  rts")
            }
            "c64" -> asmgen.out("  jsr  main.start |  lda  #31 |  sta  $01 |  rts")
            else -> asmgen.jmp("main.start")
        }
    }

    private fun memorySlabs() {
        asmgen.out("; memory slabs")
        asmgen.out("prog8_slabs\t.block")
        for((name, info) in asmgen.allMemorySlabs) {
            if(info.second>1u)
                asmgen.out("\t.align  ${info.second.toHex()}")
            asmgen.out("$name\t.fill  ${info.first}")
        }
        asmgen.out("\t.bend")
    }

    private fun footer() {
        // the global list of all floating point constants for the whole program
        asmgen.out("; global float constants")
        for (flt in asmgen.globalFloatConsts) {
            val floatFill = compTarget.machine.getFloat(flt.key).makeFloatFillAsm()
            val floatvalue = flt.key
            asmgen.out("${flt.value}\t.byte  $floatFill  ; float $floatvalue")
        }
        asmgen.out("prog8_program_end\t; end of program label for progend()")
    }

    private fun block2asm(block: Block) {
        // no longer output the initialization assignments as regular statements in the block,
        // they will be part of the prog8_init_vars init routine generated below.
        val initializers = blockVariableInitializers.getValue(block)
        val statements = block.statements.filterNot { it in initializers }

        asmgen.out("\n\n; ---- block: '${block.name}' ----")
        if(block.address!=null)
            asmgen.out("* = ${block.address!!.toHex()}")
        else {
            if("align_word" in block.options())
                asmgen.out("\t.align 2")
            else if("align_page" in block.options())
                asmgen.out("\t.align $100")
        }

        asmgen.out("${block.name}\t" + (if("force_output" in block.options()) ".block\n" else ".proc\n"))

        asmgen.outputSourceLine(block)
        zeropagevars2asm(statements, block)
        memdefs2asm(statements, block)
        vardecls2asm(statements, block)

        statements.asSequence().filterIsInstance<VarDecl>().forEach {
            if(it.type== VarDeclType.VAR && it.datatype in NumericDatatypes)
                it.value=null  // make sure every var has no init value any longer (could be set due to 'noreinit' option) because initialization is done via explicit assignment
        }

        asmgen.out("\n; subroutines in this block")

        // first translate regular statements, and then put the subroutines at the end.
        val (subroutine, stmts) = statements.partition { it is Subroutine }
        stmts.forEach { asmgen.translate(it) }
        subroutine.forEach { asmgen.translate(it) }

        if(!options.dontReinitGlobals) {
            // generate subroutine to initialize block-level (global) variables
            if (initializers.isNotEmpty()) {
                asmgen.out("prog8_init_vars\t.proc\n")
                initializers.forEach { assign -> asmgen.translate(assign) }
                asmgen.out("  rts\n  .pend")
            }
        }

        asmgen.out(if("force_output" in block.options()) "\n\t.bend\n" else "\n\t.pend\n")
    }

    internal fun translateSubroutine(sub: Subroutine) {
        var onlyVariables = false

        if(sub.inline) {
            if(options.optimize) {
                if(sub.isAsmSubroutine || callGraph.unused(sub))
                    return

                // from an inlined subroutine only the local variables are generated,
                // all other code statements are omitted in the subroutine itself
                // (they've been inlined at the call site, remember?)
                onlyVariables = true
            }
        }

        asmgen.out("")

        if(sub.isAsmSubroutine) {
            if(sub.asmAddress!=null)
                return  // already done at the memvars section

            // asmsub with most likely just an inline asm in it
            asmgen.out("${sub.name}\t.proc")
            sub.statements.forEach { asmgen.translate(it) }
            asmgen.out("  .pend\n")
        } else {
            // regular subroutine
            asmgen.out("${sub.name}\t.proc")
            zeropagevars2asm(sub.statements, null)
            memdefs2asm(sub.statements, null)

            // the main.start subroutine is the program's entrypoint and should perform some initialization logic
            if(sub.name=="start" && sub.definingBlock.name=="main")
                entrypointInitialization()

            if(functioncallAsmGen.optimizeIntArgsViaRegisters(sub)) {
                asmgen.out("; simple int arg(s) passed via register(s)")
                if(sub.parameters.size==1) {
                    val dt = sub.parameters[0].type
                    val target = AsmAssignTarget(TargetStorageKind.VARIABLE, program, asmgen, dt, sub, variableAsmName = sub.parameters[0].name)
                    if(dt in ByteDatatypes)
                        asmgen.assignRegister(RegisterOrPair.A, target)
                    else
                        asmgen.assignRegister(RegisterOrPair.AY, target)
                } else {
                    require(sub.parameters.size==2)
                    // 2 simple byte args, first in A, second in Y
                    val target1 = AsmAssignTarget(TargetStorageKind.VARIABLE, program, asmgen, sub.parameters[0].type, sub, variableAsmName = sub.parameters[0].name)
                    val target2 = AsmAssignTarget(TargetStorageKind.VARIABLE, program, asmgen, sub.parameters[1].type, sub, variableAsmName = sub.parameters[1].name)
                    asmgen.assignRegister(RegisterOrPair.A, target1)
                    asmgen.assignRegister(RegisterOrPair.Y, target2)
                }
            }

            if(!onlyVariables) {
                asmgen.out("; statements")
                sub.statements.forEach { asmgen.translate(it) }
            }

            for(removal in removals.toList()) {
                if(removal.second==sub) {
                    removal.second.remove(removal.first)
                    removals.remove(removal)
                }
            }

            asmgen.out("; variables")
            for((dt, name, addr) in sub.asmGenInfo.extraVars) {
                if(addr!=null)
                    asmgen.out("$name = $addr")
                else when(dt) {
                    DataType.UBYTE -> asmgen.out("$name    .byte  0")
                    DataType.UWORD -> asmgen.out("$name    .word  0")
                    else -> throw AssemblyError("weird dt")
                }
            }
            if(sub.asmGenInfo.usedRegsaveA)      // will probably never occur
                asmgen.out("prog8_regsaveA     .byte  0")
            if(sub.asmGenInfo.usedRegsaveX)
                asmgen.out("prog8_regsaveX     .byte  0")
            if(sub.asmGenInfo.usedRegsaveY)
                asmgen.out("prog8_regsaveY     .byte  0")
            if(sub.asmGenInfo.usedFloatEvalResultVar1)
                asmgen.out("$subroutineFloatEvalResultVar1    .byte  0,0,0,0,0")
            if(sub.asmGenInfo.usedFloatEvalResultVar2)
                asmgen.out("$subroutineFloatEvalResultVar2    .byte  0,0,0,0,0")
            vardecls2asm(sub.statements, null)
            asmgen.out("  .pend\n")
        }
    }

    private fun entrypointInitialization() {
        asmgen.out("; program startup initialization")
        asmgen.out("  cld")
        if(!options.dontReinitGlobals) {
            blockVariableInitializers.forEach {
                if (it.value.isNotEmpty())
                    asmgen.out("  jsr  ${it.key.name}.prog8_init_vars")
            }
        }

        // string and array variables in zeropage that have initializer value, should be initialized
        val stringVarsInZp = allocation.varsInZeropage.filter { it.datatype==DataType.STR && it.value!=null }
        val arrayVarsInZp = allocation.varsInZeropage.filter { it.datatype in ArrayDatatypes && it.value!=null }
        if(stringVarsInZp.isNotEmpty() || arrayVarsInZp.isNotEmpty()) {
            asmgen.out("; zp str and array initializations")
            stringVarsInZp.forEach {
                asmgen.out("""
                    lda  #<${it.name}
                    ldy  #>${it.name}
                    sta  P8ZP_SCRATCH_W1
                    sty  P8ZP_SCRATCH_W1+1
                    lda  #<${it.name}_init_value
                    ldy  #>${it.name}_init_value
                    jsr  prog8_lib.strcpy""")
            }
            arrayVarsInZp.forEach {
                val numelements = (it.value as ArrayLiteralValue).value.size
                val size = numelements * program.memsizer.memorySize(ArrayToElementTypes.getValue(it.datatype))
                asmgen.out("""
                    lda  #<${it.name}_init_value
                    ldy  #>${it.name}_init_value
                    sta  cx16.r0L
                    sty  cx16.r0H
                    lda  #<${it.name}
                    ldy  #>${it.name}
                    sta  cx16.r1L
                    sty  cx16.r1H
                    lda  #<$size
                    ldy  #>$size
                    jsr  sys.memcopy""")
            }
            asmgen.out("  jmp  +")
        }

        stringVarsInZp.forEach {
            outputStringvar(it, it.name+"_init_value")
        }
        arrayVarsInZp.forEach {
            vardecl2asm(it, it.name+"_init_value")
        }

        asmgen.out("""+       tsx
                    stx  prog8_lib.orig_stackpointer    ; required for sys.exit()                    
                    ldx  #255       ; init estack ptr
                    clv
                    clc""")
    }

    private fun zeropagevars2asm(statements: List<Statement>, inBlock: Block?) {
        asmgen.out("; vars allocated on zeropage")
        val variables = statements.asSequence().filterIsInstance<VarDecl>().filter { it.type==VarDeclType.VAR }
        val blockname = inBlock?.name
        for(variable in variables) {
            if(blockname=="prog8_lib" && variable.name.startsWith("P8ZP_SCRATCH_"))
                continue       // the "hooks" to the temp vars are not generated as new variables
            val scopedName = variable.scopedName
            val zpAlloc = asmgen.zeropage.variables[scopedName]
            if (zpAlloc == null) {
                // This var is not on the ZP yet. Attempt to move it there if it's an integer type
                if(variable.zeropage != ZeropageWish.NOT_IN_ZEROPAGE &&
                    variable.datatype in IntegerDatatypes
                    && options.zeropage != ZeropageType.DONTUSE) {
                    val result = asmgen.zeropage.allocate(scopedName, variable.datatype, null, null, errors)
                    errors.report()
                    result.fold(
                        success = { (address, _) -> asmgen.out("${variable.name} = $address\t; zp ${variable.datatype}") },
                        failure = { /* leave it as it is, not on zeropage. */ }
                    )
                }
            } else {
                // Var has been placed in ZP, just output the address
                val lenspec = when(zpAlloc.second.first) {
                    DataType.FLOAT,
                    DataType.STR,
                    in ArrayDatatypes -> " ${zpAlloc.second.second} bytes"
                    else -> ""
                }
                asmgen.out("${variable.name} = ${zpAlloc.first}\t; zp ${variable.datatype} $lenspec")
            }
        }
    }

    private fun vardecl2asm(decl: VarDecl, nameOverride: String?=null) {
        val name = nameOverride ?: decl.name
        val value = decl.value
        val staticValue: Number =
            if(value!=null) {
                if(value is NumericLiteralValue) {
                    if(value.type== DataType.FLOAT)
                        value.number
                    else
                        value.number.toInt()
                } else {
                    if(decl.datatype in NumericDatatypes)
                        throw AssemblyError("can only deal with constant numeric values for global vars $value at ${decl.position}")
                    else 0
                }
            } else 0

        when (decl.datatype) {
            DataType.UBYTE -> asmgen.out("$name\t.byte  ${staticValue.toHex()}")
            DataType.BYTE -> asmgen.out("$name\t.char  $staticValue")
            DataType.UWORD -> asmgen.out("$name\t.word  ${staticValue.toHex()}")
            DataType.WORD -> asmgen.out("$name\t.sint  $staticValue")
            DataType.FLOAT -> {
                if(staticValue==0) {
                    asmgen.out("$name\t.byte  0,0,0,0,0  ; float")
                } else {
                    val floatFill = compTarget.machine.getFloat(staticValue).makeFloatFillAsm()
                    asmgen.out("$name\t.byte  $floatFill  ; float $staticValue")
                }
            }
            DataType.STR -> {
                throw AssemblyError("all string vars should have been interned into prog")
            }
            DataType.ARRAY_UB -> {
                val data = makeArrayFillDataUnsigned(decl)
                if (data.size <= 16)
                    asmgen.out("$name\t.byte  ${data.joinToString()}")
                else {
                    asmgen.out(name)
                    for (chunk in data.chunked(16))
                        asmgen.out("  .byte  " + chunk.joinToString())
                }
            }
            DataType.ARRAY_B -> {
                val data = makeArrayFillDataSigned(decl)
                if (data.size <= 16)
                    asmgen.out("$name\t.char  ${data.joinToString()}")
                else {
                    asmgen.out(name)
                    for (chunk in data.chunked(16))
                        asmgen.out("  .char  " + chunk.joinToString())
                }
            }
            DataType.ARRAY_UW -> {
                val data = makeArrayFillDataUnsigned(decl)
                if (data.size <= 16)
                    asmgen.out("$name\t.word  ${data.joinToString()}")
                else {
                    asmgen.out(name)
                    for (chunk in data.chunked(16))
                        asmgen.out("  .word  " + chunk.joinToString())
                }
            }
            DataType.ARRAY_W -> {
                val data = makeArrayFillDataSigned(decl)
                if (data.size <= 16)
                    asmgen.out("$name\t.sint  ${data.joinToString()}")
                else {
                    asmgen.out(name)
                    for (chunk in data.chunked(16))
                        asmgen.out("  .sint  " + chunk.joinToString())
                }
            }
            DataType.ARRAY_F -> {
                val array =
                    if(decl.value!=null)
                        (decl.value as ArrayLiteralValue).value
                    else {
                        // no init value, use zeros
                        val zero = decl.zeroElementValue()
                        Array(decl.arraysize!!.constIndex()!!) { zero }
                    }
                val floatFills = array.map {
                    val number = (it as NumericLiteralValue).number
                    compTarget.machine.getFloat(number).makeFloatFillAsm()
                }
                asmgen.out(name)
                for (f in array.zip(floatFills))
                    asmgen.out("  .byte  ${f.second}  ; float ${f.first}")
            }
            else -> {
                throw AssemblyError("weird dt")
            }
        }
    }

    private fun memdefs2asm(statements: List<Statement>, inBlock: Block?) {
        val blockname = inBlock?.name

        asmgen.out("\n; memdefs and kernal subroutines")
        val memvars = statements.asSequence().filterIsInstance<VarDecl>().filter { it.type==VarDeclType.MEMORY || it.type==VarDeclType.CONST }
        for(m in memvars) {
            if(blockname!="prog8_lib" || !m.name.startsWith("P8ZP_SCRATCH_"))      // the "hooks" to the temp vars are not generated as new variables
                if(m.value is NumericLiteralValue)
                    asmgen.out("  ${m.name} = ${(m.value as NumericLiteralValue).number.toHex()}")
                else
                    asmgen.out("  ${m.name} = ${asmgen.asmVariableName((m.value as AddressOf).identifier)}")
        }
        val asmSubs = statements.asSequence().filterIsInstance<Subroutine>().filter { it.isAsmSubroutine }
        for(sub in asmSubs) {
            val addr = sub.asmAddress
            if(addr!=null) {
                if(sub.statements.isNotEmpty())
                    throw AssemblyError("kernal subroutine cannot have statements")
                asmgen.out("  ${sub.name} = ${addr.toHex()}")
            }
        }
    }

    private fun vardecls2asm(statements: List<Statement>, inBlock: Block?) {
        asmgen.out("\n; non-zeropage variables")
        val vars = statements.asSequence()
            .filterIsInstance<VarDecl>()
            .filter {
                it.type==VarDeclType.VAR
                        && it.zeropage!= ZeropageWish.REQUIRE_ZEROPAGE
                        && it.scopedName !in asmgen.zeropage.variables
            }

        vars.filter { it.datatype == DataType.STR && shouldActuallyOutputStringVar(it) }
            .forEach { outputStringvar(it) }

        // non-string variables
        val blockname = inBlock?.name

        vars.filter{ it.datatype != DataType.STR }.sortedBy { it.datatype }.forEach {
            require(it.zeropage!= ZeropageWish.REQUIRE_ZEROPAGE)
            if(!asmgen.isZpVar(it.scopedName)) {
                if(blockname!="prog8_lib" || !it.name.startsWith("P8ZP_SCRATCH_"))      // the "hooks" to the temp vars are not generated as new variables
                    vardecl2asm(it)
            }
        }
    }

    private fun shouldActuallyOutputStringVar(strvar: VarDecl): Boolean {
        if(strvar.sharedWithAsm)
            return true
        val uses = callGraph.usages(strvar)
        val onlyInMemoryFuncs = uses.all {
            val builtinfunc = (it.parent as? IFunctionCall)?.target?.targetStatement(program) as? BuiltinFunctionPlaceholder
            builtinfunc?.name=="memory"
        }
        return !onlyInMemoryFuncs
    }

    private fun outputStringvar(strdecl: VarDecl, nameOverride: String?=null) {
        val varname = nameOverride ?: strdecl.name
        val sv = strdecl.value as StringLiteralValue
        asmgen.out("$varname\t; ${strdecl.datatype} ${sv.encoding}:\"${escape(sv.value).replace("\u0000", "<NULL>")}\"")
        val bytes = compTarget.encodeString(sv.value, sv.encoding).plus(0.toUByte())
        val outputBytes = bytes.map { "$" + it.toString(16).padStart(2, '0') }
        for (chunk in outputBytes.chunked(16))
            asmgen.out("  .byte  " + chunk.joinToString())
    }

    private fun makeArrayFillDataUnsigned(decl: VarDecl): List<String> {
        val array =
            if(decl.value!=null)
                (decl.value as ArrayLiteralValue).value
            else {
                // no array init value specified, use a list of zeros
                val zero = decl.zeroElementValue()
                Array(decl.arraysize!!.constIndex()!!) { zero }
            }
        return when (decl.datatype) {
            DataType.ARRAY_UB ->
                // byte array can never contain pointer-to types, so treat values as all integers
                array.map {
                    val number = (it as NumericLiteralValue).number.toInt()
                    "$"+number.toString(16).padStart(2, '0')
                }
            DataType.ARRAY_UW -> array.map {
                when (it) {
                    is NumericLiteralValue -> {
                        "$" + it.number.toInt().toString(16).padStart(4, '0')
                    }
                    is AddressOf -> {
                        asmgen.asmSymbolName(it.identifier)
                    }
                    is IdentifierReference -> {
                        asmgen.asmSymbolName(it)
                    }
                    else -> throw AssemblyError("weird array elt dt")
                }
            }
            else -> throw AssemblyError("invalid arraysize type")
        }
    }

    private fun makeArrayFillDataSigned(decl: VarDecl): List<String> {
        val array =
            if(decl.value!=null) {
                if(decl.value !is ArrayLiteralValue)
                    throw AssemblyError("can only use array literal values as array initializer value")
                (decl.value as ArrayLiteralValue).value
            }
            else {
                // no array init value specified, use a list of zeros
                val zero = decl.zeroElementValue()
                Array(decl.arraysize!!.constIndex()!!) { zero }
            }
        return when (decl.datatype) {
            DataType.ARRAY_UB ->
                // byte array can never contain pointer-to types, so treat values as all integers
                array.map {
                    val number = (it as NumericLiteralValue).number.toInt()
                    "$"+number.toString(16).padStart(2, '0')
                }
            DataType.ARRAY_B ->
                // byte array can never contain pointer-to types, so treat values as all integers
                array.map {
                    val number = (it as NumericLiteralValue).number.toInt()
                    val hexnum = number.absoluteValue.toString(16).padStart(2, '0')
                    if(number>=0)
                        "$$hexnum"
                    else
                        "-$$hexnum"
                }
            DataType.ARRAY_UW -> array.map {
                val number = (it as NumericLiteralValue).number.toInt()
                "$" + number.toString(16).padStart(4, '0')
            }
            DataType.ARRAY_W -> array.map {
                val number = (it as NumericLiteralValue).number.toInt()
                val hexnum = number.absoluteValue.toString(16).padStart(4, '0')
                if(number>=0)
                    "$$hexnum"
                else
                    "-$$hexnum"
            }
            else -> throw AssemblyError("invalid arraysize type ${decl.datatype}")
        }
    }

}